#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import random
import re

from gurux_dlms.objects.GXDLMSObject import GXDLMSObject

from smartmeter_datacollector.smartmeter.cosem import Cosem
from smartmeter_datacollector.smartmeter.hdlc_dlms_parser import HdlcDlmsParser

from .utils import *


class TestHdlcParserUnencrypted:
    def test_extract_hdlc_data_framewise(self, unencrypted_valid_data_lg: List[bytes]):
        parser = HdlcDlmsParser(Cosem("", []))

        for frame in unencrypted_valid_data_lg:
            assert not parser.extract_data_from_hdlc_frames()
            parser.append_to_hdlc_buffer(frame)

        assert parser.extract_data_from_hdlc_frames()

    def test_extract_hdlc_data_in_halfframes(self, unencrypted_valid_data_lg: List[bytes]):
        parser = HdlcDlmsParser(Cosem("", []))

        for frame in unencrypted_valid_data_lg:
            frame: bytes
            assert not parser.extract_data_from_hdlc_frames()
            parser.append_to_hdlc_buffer(frame[:int(len(frame)/2)])
            assert not parser.extract_data_from_hdlc_frames()
            parser.append_to_hdlc_buffer(frame[int(len(frame)/2):])

        assert parser.extract_data_from_hdlc_frames()


class TestDlmsParserUnencrypted:
    def test_hdlc_to_dlms_objects_with_pushlist(self, unencrypted_valid_data_lg: List[bytes], cosem_config_lg: Cosem):
        parser = prepare_parser(unencrypted_valid_data_lg, cosem_config_lg)
        dlms_objects = parser.parse_to_dlms_objects()

        assert isinstance(dlms_objects, list)
        assert len(dlms_objects) == 16
        obis_pattern = re.compile(r"(\d+\.){5}\d+")
        for obj in dlms_objects:
            assert isinstance(obj, GXDLMSObject)
            assert re.match(obis_pattern, str(obj.logicalName))

    def test_parse_dlms_to_meter_data(self, unencrypted_valid_data_lg: List[bytes], cosem_config_lg: Cosem):
        parser = prepare_parser(unencrypted_valid_data_lg, cosem_config_lg)
        dlms_objects = parser.parse_to_dlms_objects()
        meter_data = parser.convert_dlms_bundle_to_reader_data(dlms_objects)

        assert isinstance(meter_data, list)
        assert len(meter_data) == 11
        assert any(data.type == MeterDataPointTypes.ACTIVE_POWER_P.value for data in meter_data)
        assert any(data.type == MeterDataPointTypes.ACTIVE_POWER_N.value for data in meter_data)
        assert all(isinstance(data.value, float) for data in meter_data)
        assert all(data.source == "LGZ1030655933512" for data in meter_data)
        assert all(data.timestamp.strftime(r"%m/%d/%y %H:%M:%S") == "07/06/21 14:58:18" for data in meter_data)

    def test_parse_dlms_to_meter_data2(self, unencrypted_valid_data_lg2: List[bytes], cosem_config_lg: Cosem):
        parser = prepare_parser(unencrypted_valid_data_lg2, cosem_config_lg)
        dlms_objects = parser.parse_to_dlms_objects()
        meter_data = parser.convert_dlms_bundle_to_reader_data(dlms_objects)

        assert isinstance(meter_data, list)
        assert len(meter_data) == 8

    def test_ignore_not_parsable_data_to_meter_data(self, unencrypted_invalid_data_lg: List[bytes], cosem_config_lg: Cosem):
        parser = prepare_parser(unencrypted_invalid_data_lg, cosem_config_lg)
        dlms_objects = parser.parse_to_dlms_objects()
        meter_data = parser.convert_dlms_bundle_to_reader_data(dlms_objects)

        assert isinstance(meter_data, list)
        assert len(meter_data) == 5


class TestDlmsParserEncrypted:
    def test_hdlc_to_dlms_objects_without_pushlist(self, encrypted_data_no_pushlist_lg: List[bytes], cosem_config_lg: Cosem):
        parser = prepare_parser(encrypted_data_no_pushlist_lg, cosem_config_lg, "F08660A6C19D2048556BF623AB0257E6")
        dlms_objects = parser.parse_to_dlms_objects()

        assert isinstance(dlms_objects, list)
        assert len(dlms_objects) == 16
        obis_pattern = re.compile(r"(\d+\.){5}\d+")
        for obj in dlms_objects:
            assert isinstance(obj, GXDLMSObject)
            assert re.match(obis_pattern, str(obj.logicalName))



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import json
from datetime import datetime, timezone

from smartmeter_datacollector.smartmeter.meter_data import MeterDataPoint, MeterDataPointType


def test_meter_data_serialize():
    test_type = MeterDataPointType("TEST_TYPE", "test type", "unit")
    data_point = MeterDataPoint(test_type, 1.0, "TestSource", datetime.now(timezone.utc))

    data_point_json = data_point.to_json()

    result = json.loads(data_point_json)
    assert result['type']['identifier'] == data_point.type.identifier
    assert result['type']['name'] == data_point.type.name
    assert result['type']['unit'] == data_point.type.unit
    assert result['value'] == data_point.value
    assert result['source'] == data_point.source
    assert result['timestamp'] == data_point.timestamp.isoformat()



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import asyncio
import sys
from datetime import datetime, timezone

import pytest
from pytest_mock import MockerFixture

from smartmeter_datacollector.collector import Collector
from smartmeter_datacollector.sinks.data_sink import DataSink
from smartmeter_datacollector.smartmeter.meter_data import MeterDataPoint, MeterDataPointType


@pytest.fixture
def test_type() -> MeterDataPointType:
    return MeterDataPointType("TEST_TYPE", "test type", "unit")


@pytest.mark.asyncio
async def test_collector_with_one_sink(mocker: MockerFixture, test_type: MeterDataPointType):
    coll = Collector()
    sink = mocker.AsyncMock(DataSink)
    data_point = MeterDataPoint(test_type, 0.0, "test_source", datetime.now(timezone.utc))

    coll.register_sink(sink)
    coll.notify([data_point])
    routine = coll.process_queue()

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(routine, 0.1)

    sink.send.assert_awaited_once_with(data_point)


@pytest.mark.asyncio
async def test_collector_with_one_sink_multiple_data_points(mocker: MockerFixture, test_type: MeterDataPointType):
    coll = Collector()
    sink = mocker.AsyncMock(DataSink)

    coll.register_sink(sink)
    point0 = MeterDataPoint(test_type, 0.0, "test_source", datetime.now(timezone.utc))
    point1 = MeterDataPoint(test_type, 1.0, "test_source", datetime.now(timezone.utc))
    coll.notify([point0, point1])
    routine = coll.process_queue()

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(routine, 0.1)

    sink.send.assert_awaited_with(point1)
    assert sink.send.await_count == 2


@pytest.mark.asyncio
async def test_collector_with_two_sinks(mocker: MockerFixture, test_type: MeterDataPointType):
    coll = Collector()
    sink0 = mocker.AsyncMock(DataSink)
    sink1 = mocker.AsyncMock(DataSink)

    data_point = MeterDataPoint(test_type, 0.0, "test_source", datetime.now(timezone.utc))

    coll.register_sink(sink0)
    coll.register_sink(sink1)
    coll.notify([data_point])
    routine = coll.process_queue()

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(routine, 0.1)

    sink0.send.assert_awaited_once_with(data_point)
    sink1.send.assert_awaited_once_with(data_point)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from abc import ABC, abstractmethod
from typing import Callable


class ReaderError(Exception):
    pass


# pylint: disable=too-few-public-methods
class Reader(ABC):
    def __init__(self, callback: Callable[[bytes], None]) -> None:
        self._callback = callback

    @abstractmethod
    async def start_and_listen(self) -> None:
        pass



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from smartmeter_datacollector.smartmeter.obis import OBISCode


def test_create_from_gurux_string():
    obis_string = "1.0.1.7.0.255"
    obis = OBISCode.from_string(obis_string)
    assert obis
    assert obis.a == 1
    assert obis.b == 0
    assert obis.c == 1
    assert obis.d == 7
    assert obis.e == 0
    assert obis.f == 255


def test_create_from_IEC_62056_string():
    obis_string = "1-0:1.7.0*255"
    obis = OBISCode.from_string(obis_string)
    assert obis
    assert obis.a == 1
    assert obis.b == 0
    assert obis.c == 1
    assert obis.d == 7
    assert obis.e == 0
    assert obis.f == 255


def test_create_from_bytes():
    obis_bytes = bytes.fromhex("01 01 01 08 02 FF")
    obis = OBISCode.from_bytes(obis_bytes)
    assert obis
    assert obis.a == 1
    assert obis.b == 1
    assert obis.c == 1
    assert obis.d == 8
    assert obis.e == 2
    assert obis.f == 255


def test_to_IEC62056_string():
    obis = OBISCode(1, 0, 1, 7, 0, 255)
    assert str(obis) == "1-0:1.7.0*255"


def test_to_gurux_string():
    obis = OBISCode(1, 0, 1, 7, 0, 255)
    assert obis.to_gurux_str() == "1.0.1.7.0.255"


def test_only_compare_relevant_parts():
    obis = OBISCode(1, 0, 1, 7, 0, 255)
    assert obis == OBISCode(1, 1, 1, 7, 0, 255)
    assert obis == OBISCode(1, 0, 1, 7, 0, 128)
    assert obis != OBISCode(0, 0, 1, 7, 0, 255)
    assert obis != OBISCode(1, 0, 2, 7, 0, 255)
    assert obis != OBISCode(1, 0, 1, 16, 0, 255)
    assert obis != OBISCode(1, 0, 1, 7, 1, 255)


def test_hash_of_obis():
    obis = OBISCode(1, 0, 1, 7, 0, 255)
    obis_diff = OBISCode(2, 0, 1, 7, 0, 255)
    obis_same = OBISCode(1, 2, 1, 7, 0, 255)

    assert hash(obis) == hash(obis_same)
    assert hash(obis) != hash(obis_diff)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from smartmeter_datacollector.app import main

main()



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import sys
from typing import List

import pytest
from pytest_mock.plugin import MockerFixture

from smartmeter_datacollector.smartmeter.lge450 import LGE450
from smartmeter_datacollector.smartmeter.meter_data import MeterDataPointTypes

from .utils import *


@pytest.mark.asyncio
async def test_lge450_initialization(mocker: MockerFixture):
    observer = mocker.stub()
    test_bytes = bytes([1, 2, 3])
    serial_mock = mocker.patch("smartmeter_datacollector.smartmeter.meter.SerialReader",
                               autospec=True).return_value
    meter = LGE450("/test/port")
    serial_mock.start_and_listen.side_effect = meter._data_received(test_bytes)
    meter.register(observer)
    await meter.start()

    serial_mock.start_and_listen.assert_awaited_once()
    observer.assert_not_called


@pytest.mark.asyncio
async def test_lge450_parse_and_provide_unencrypted_data(mocker: MockerFixture,
                                                         unencrypted_valid_data_lg: List[bytes]):
    observer = mocker.stub("collector_mock")
    observer.mock_add_spec(['notify'])
    serial_mock = mocker.patch("smartmeter_datacollector.smartmeter.meter.SerialReader",
                               autospec=True).return_value
    meter = LGE450("/test/port")
    meter.register(observer)

    def data_received():
        for frame in unencrypted_valid_data_lg:
            meter._data_received(frame)
    serial_mock.start_and_listen.side_effect = data_received

    await meter.start()

    serial_mock.start_and_listen.assert_awaited_once()
    observer.notify.assert_called_once()
    values = observer.notify.call_args.args[0]
    assert isinstance(values, list)
    assert any(data.type == MeterDataPointTypes.ACTIVE_POWER_P.value for data in values)
    assert any(data.type == MeterDataPointTypes.ACTIVE_POWER_N.value for data in values)
    assert any(data.type == MeterDataPointTypes.REACTIVE_POWER_P.value for data in values)
    assert any(data.type == MeterDataPointTypes.REACTIVE_POWER_N.value for data in values)
    assert any(data.type == MeterDataPointTypes.ACTIVE_ENERGY_P.value for data in values)
    assert any(data.type == MeterDataPointTypes.ACTIVE_ENERGY_N.value for data in values)
    assert any(data.type == MeterDataPointTypes.REACTIVE_ENERGY_Q1.value for data in values)
    assert any(data.type == MeterDataPointTypes.REACTIVE_ENERGY_Q2.value for data in values)
    assert any(data.type == MeterDataPointTypes.REACTIVE_ENERGY_Q3.value for data in values)
    assert any(data.type == MeterDataPointTypes.REACTIVE_ENERGY_Q4.value for data in values)
    assert any(data.type == MeterDataPointTypes.POWER_FACTOR.value for data in values)
    assert all(data.source == "LGZ1030655933512" for data in values)
    assert all(data.timestamp.strftime(r"%m/%d/%y %H:%M:%S") == "07/06/21 14:58:18" for data in values)


@pytest.mark.asyncio
async def test_lge450_do_not_provide_invalid_data(mocker: MockerFixture,
                                                  unencrypted_invalid_data_lg: List[bytes]):
    observer = mocker.stub("collector_mock")
    observer.mock_add_spec(['notify'])
    serial_mock = mocker.patch("smartmeter_datacollector.smartmeter.meter.SerialReader",
                               autospec=True).return_value
    meter = LGE450("/test/port")
    meter.register(observer)

    def data_received():
        for frame in unencrypted_invalid_data_lg:
            meter._data_received(frame)
    serial_mock.start_and_listen.side_effect = data_received

    await meter.start()

    serial_mock.start_and_listen.assert_awaited_once()
    observer.notify.assert_not_called



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging

from ..smartmeter.meter_data import MeterDataPoint
from .data_sink import DataSink


class LoggerSink(DataSink):
    def __init__(self, logger_name: str) -> None:
        self._logger = logging.getLogger(logger_name)
        self._logger.setLevel(logging.INFO)

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass

    async def send(self, data_point: MeterDataPoint) -> None:
        self._logger.info(str(data_point))



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import re
from dataclasses import dataclass, field
from typing import ClassVar, Union

REGEX = r"^(\d{1,3})\W(\d{1,3})\W(\d{1,3})\W(\d{1,3})\W(\d{1,3})\W(\d{1,3})$"


@dataclass(frozen=True)
class OBISCode:
    PATTERN: ClassVar[re.Pattern] = re.compile(REGEX)

    # pylint: disable=invalid-name
    a: int
    b: int = field(compare=False)
    c: int
    d: int
    e: int
    f: int = field(default=255, compare=False)

    def __str__(self) -> str:
        return f"{self.a}-{self.b}:{self.c}.{self.d}.{self.e}*{self.f}"

    def to_gurux_str(self) -> str:
        return f"{self.a}.{self.b}.{self.c}.{self.d}.{self.e}.{self.f}"

    def is_same_type(self, other: 'OBISCode') -> bool:
        """Compares only A, C, D, E parts of an OBIS code."""
        return (self.a == other.a and
                self.b == other.b and
                self.c == other.c and
                self.d == other.d)

    @classmethod
    def from_string(cls, obis_string: str) -> 'OBISCode':
        match = cls.PATTERN.match(obis_string)
        if not match:
            raise ValueError(f"Invalid OBIS string {obis_string}.")
        groups = match.groups()
        return cls(*(int(g) for g in groups))

    @classmethod
    def from_bytes(cls, obis_bytes: Union[bytes, bytearray]) -> 'OBISCode':
        if not cls.is_obis(obis_bytes):
            raise ValueError("Invalid OBIS bytes.")
        return cls(*obis_bytes)

    @staticmethod
    def is_obis(data: Union[bytes, bytearray]) -> bool:
        return (len(data) == 6 and
                data[0] < 10 and
                data[1] <= 64 and
                data[3] < 128)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import asyncio
import logging
from asyncio import QueueFull
from typing import List

from .sinks.data_sink import DataSink
from .smartmeter.meter_data import MeterDataPoint

LOGGER = logging.getLogger("collector")


class Collector:
    def __init__(self) -> None:
        self._queue = asyncio.Queue()
        self._data_sinks: List[DataSink] = []

    def register_sink(self, sink: DataSink) -> None:
        assert isinstance(sink, DataSink)
        self._data_sinks.append(sink)

    def notify(self, reader_data_points: List[MeterDataPoint]) -> None:
        for point in reader_data_points:
            try:
                self._queue.put_nowait(point)
            except QueueFull:
                LOGGER.warning("Queue is full. Current data points are dropped.")
                return

    async def process_queue(self) -> None:
        while True:
            data_point: MeterDataPoint = await self._queue.get()
            for sink in self._data_sinks:
                await sink.send(data_point)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import json
import logging
import ssl
from configparser import SectionProxy
from dataclasses import dataclass
from typing import Optional

from asyncio_mqtt import Client
from asyncio_mqtt.client import ProtocolVersion
from asyncio_mqtt.error import MqttCodeError, MqttError
from paho.mqtt.client import MQTT_ERR_NO_CONN

from ..smartmeter.meter_data import MeterDataPoint
from .data_sink import DataSink

LOGGER = logging.getLogger("sink")


# pylint: disable=too-many-instance-attributes
@dataclass
class MqttConfig:
    broker_host: str
    port: int = 1883
    use_tls: bool = False
    username: Optional[str] = None
    password: Optional[str] = None
    client_id: Optional[str] = None
    ca_cert_path: Optional[str] = None
    check_hostname: bool = True
    client_cert_path: Optional[str] = None
    client_key_path: Optional[str] = None

    def with_tls(self, ca_cert_path: Optional[str] = None, check_hostname: bool = True) -> "MqttConfig":
        self.use_tls = True
        self.ca_cert_path = ca_cert_path
        self.check_hostname = check_hostname
        return self

    def with_user_pass_auth(self, username: str, password: str) -> "MqttConfig":
        self.username = username
        self.password = password
        return self

    def with_client_cert_auth(self, cert_path: str, key_path: str) -> "MqttConfig":
        self.use_tls = True
        self.client_cert_path = cert_path
        self.client_key_path = key_path
        return self

    @staticmethod
    def from_sink_config(config: SectionProxy) -> "MqttConfig":
        mqtt_cfg = MqttConfig(config.get("host"), config.getint("port", 1883))
        if config.getboolean("tls", fallback=False):
            mqtt_cfg.with_tls(config.get("ca_file_path"), config.getboolean("check_hostname", True))
        if config.get("username") and config.get("password"):
            mqtt_cfg.with_user_pass_auth(
                config.get("username"),
                config.get("password"))
        if config.get("client_cert_path") and config.get("client_key_path"):
            mqtt_cfg.with_client_cert_auth(
                config.get("client_cert_path"),
                config.get("client_key_path")
            )
        return mqtt_cfg


class MqttDataSink(DataSink):
    TIMEOUT = 3

    def __init__(self, config: MqttConfig) -> None:
        tls_context = None
        if config.use_tls:
            tls_context = self._build_ssl_context(
                config.ca_cert_path,
                config.check_hostname,
                config.client_cert_path,
                config.client_key_path)

        user_pass_auth = {}
        if config.username and config.password:
            user_pass_auth["username"] = config.username
            user_pass_auth["password"] = config.password

        self._client = Client(
            hostname=config.broker_host,
            port=config.port,
            client_id=config.client_id,
            tls_context=tls_context,
            protocol=ProtocolVersion.V311,
            clean_session=True,
            **user_pass_auth)

    @staticmethod
    def _build_ssl_context(
        ca_file_path: Optional[str] = None,
        check_hostname: bool = True,
        client_cert_path: Optional[str] = None,
        client_key_path: Optional[str] = None
    ) -> ssl.SSLContext:
        context = ssl.create_default_context(cafile=ca_file_path)

        if ca_file_path:
            context.check_hostname = check_hostname

        if client_cert_path:
            try:
                context.load_cert_chain(client_cert_path, client_key_path, None)
            except ssl.SSLError as ex:
                LOGGER.error("Client certificate does not match with the key and is ignored. '%s'", ex)
        return context

    async def start(self) -> None:
        if await self._connect_to_server():
            LOGGER.info("Connected to MQTT broker.")

    async def stop(self) -> None:
        await self._disconnect_from_server()
        LOGGER.info("Disconnected from MQTT broker.")

    async def send(self, data_point: MeterDataPoint) -> None:
        topic = MqttDataSink.get_topic_name_for_datapoint(data_point)
        dp_json = self.data_point_to_mqtt_json(data_point)
        try:
            await self._client.publish(topic, dp_json)
            LOGGER.debug("%s sent to MQTT broker.", dp_json)
        except ValueError as ex:
            LOGGER.error("MQTT payload or topic is invalid: '%s'", ex)
        except MqttCodeError as ex:
            if ex.rc == MQTT_ERR_NO_CONN:
                if await self._connect_to_server():
                    LOGGER.info("Reconnected to MQTT broker.")
                    try:
                        await self._client.publish(topic, dp_json)
                        LOGGER.debug("%s sent to MQTT broker.", dp_json)
                    except MqttError:
                        LOGGER.warning("MQTT message not sent.")
            else:
                LOGGER.error("MQTT message sending error: '%s'", ex)
        except MqttError as ex:
            LOGGER.error("MQTT message sending error: '%s'", ex)

    async def _connect_to_server(self) -> bool:
        try:
            await self._client.connect(timeout=self.TIMEOUT)
        except MqttError as ex:
            LOGGER.error(ex)
            return False
        return True

    async def _disconnect_from_server(self) -> None:
        try:
            await self._client.disconnect(timeout=self.TIMEOUT)
        except MqttError as ex:
            LOGGER.error(ex)
            await self._client.force_disconnect()

    @staticmethod
    def get_topic_name_for_datapoint(data_point: MeterDataPoint) -> str:
        return f"smartmeter/{data_point.source}/{data_point.type.identifier}"

    @staticmethod
    def data_point_to_mqtt_json(data_point: MeterDataPoint) -> str:
        return json.dumps({
            "value": data_point.value,
            "timestamp": int(data_point.timestamp.timestamp())
        })



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
from typing import Optional

import serial

from .cosem import Cosem
from .meter import MeterError, SerialHdlcDlmsMeter
from .reader import ReaderError
from .serial_reader import SerialConfig

LOGGER = logging.getLogger("smartmeter")


class LGE570(SerialHdlcDlmsMeter):
    BAUDRATE = 2400

    def __init__(self, port: str,
                 baudrate: int = BAUDRATE,
                 decryption_key: Optional[str] = None,
                 use_system_time: bool = False) -> None:
        serial_config = SerialConfig(
            port=port,
            baudrate=baudrate,
            data_bits=serial.EIGHTBITS,
            parity=serial.PARITY_EVEN,
            stop_bits=serial.STOPBITS_ONE,
            termination=SerialHdlcDlmsMeter.HDLC_FLAG
        )
        cosem = Cosem(fallback_id=port)
        try:
            super().__init__(serial_config, cosem, decryption_key, use_system_time)
        except ReaderError as ex:
            LOGGER.fatal("Unable to setup serial reader for L+G E570. '%s'", ex)
            raise MeterError("Failed setting up L+G E570.") from ex

        LOGGER.info("Successfully set up L+G E570 smart meter on '%s'.", port)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
from configparser import ConfigParser
from typing import List

from .collector import Collector
from .config import InvalidConfigError
from .sinks.data_sink import DataSink
from .sinks.logger_sink import LoggerSink
from .sinks.mqtt_sink import MqttConfig, MqttDataSink
from .smartmeter.iskraam550 import IskraAM550
from .smartmeter.kamstrup_han import KamstrupHAN
from .smartmeter.lge360 import LGE360
from .smartmeter.lge450 import LGE450
from .smartmeter.lge570 import LGE570
from .smartmeter.meter import Meter, MeterError


def build_meters(config: ConfigParser) -> List[Meter]:
    meters = []
    for section_name in filter(lambda sec: sec.startswith("reader"), config.sections()):
        meter_config = config[section_name]
        meter_type = meter_config.get('type')
        try:
            if meter_type == "lge450":
                meters.append(LGE450(
                    port=meter_config.get('port', "/dev/ttyUSB0"),
                    baudrate=meter_config.getint('baudrate', LGE450.BAUDRATE),
                    decryption_key=meter_config.get('key'),
                    use_system_time=meter_config.getboolean('systemtime', False)
                ))
            elif meter_type == "lge570":
                meters.append(LGE570(
                    port=meter_config.get('port', "/dev/ttyUSB0"),
                    baudrate=meter_config.getint('baudrate', LGE570.BAUDRATE),
                    decryption_key=meter_config.get('key'),
                    use_system_time=meter_config.getboolean('systemtime', False)
                ))
            elif meter_type == "lge360":
                meters.append(LGE360(
                    port=meter_config.get('port', "/dev/ttyUSB0"),
                    baudrate=meter_config.getint('baudrate', LGE360.BAUDRATE),
                    decryption_key=meter_config.get('key'),
                    use_system_time=meter_config.getboolean('systemtime', False)
                ))
            elif meter_type == "iskraam550":
                meters.append(IskraAM550(
                    port=meter_config.get('port', "/dev/ttyUSB0"),
                    baudrate=meter_config.getint('baudrate', IskraAM550.BAUDRATE),
                    decryption_key=meter_config.get('key'),
                    use_system_time=meter_config.getboolean('systemtime', False)
                ))
            elif meter_type == "kamstrup_han":
                meters.append(KamstrupHAN(
                    port=meter_config.get('port', "/dev/ttyUSB0"),
                    baudrate=meter_config.getint('baudrate', KamstrupHAN.BAUDRATE),
                    decryption_key=meter_config.get('key'),
                    use_system_time=meter_config.getboolean('systemtime', False)
                ))
            else:
                raise InvalidConfigError(f"'type' is invalid or missing: {meter_type}")
        except MeterError as ex:
            logging.warning("%s Skipping smart meter.", ex)
            continue
    return meters


def build_sinks(config: ConfigParser) -> List[DataSink]:
    sinks = []
    for section_name in filter(lambda sec: sec.startswith("sink"), config.sections()):
        sink_config = config[section_name]
        sink_type = sink_config.get('type')

        if sink_type == "logger":
            sinks.append(LoggerSink(
                logger_name=sink_config.get('name', "DataLogger")
            ))
        elif sink_type == "mqtt":
            mqtt_config = MqttConfig.from_sink_config(sink_config)
            sinks.append(MqttDataSink(mqtt_config))
        else:
            raise InvalidConfigError(f"'type' is invalid or missing: {sink_type}")
    return sinks


def build_collector(readers: List[Meter], sinks: List[DataSink]) -> Collector:
    collector = Collector()

    for sink in sinks:
        collector.register_sink(sink)
    for reader in readers:
        reader.register(collector)
    return collector
